from typing import Any, Dict, List, Type, TypeVar, cast

import attr

from ..models.workflow_node_type import WorkflowNodeType

T = TypeVar("T", bound="WorkflowNode")


@attr.s(auto_attribs=True)
class WorkflowNode:
    """
    Attributes:
        type (WorkflowNodeType):
        id (str):
        depends (List[str]):
    """

    type: WorkflowNodeType
    id: str
    depends: List[str]
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        type = self.type.value

        id = self.id
        depends = self.depends

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "type": type,
                "id": id,
                "depends": depends,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        type = WorkflowNodeType(d.pop("type"))

        id = d.pop("id")

        depends = cast(List[str], d.pop("depends"))

        workflow_node = cls(
            type=type,
            id=id,
            depends=depends,
        )

        workflow_node.additional_properties = d
        return workflow_node

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import argparse
import fcntl
import json
import os
import signal
import struct
import sys
import termios
import tty
from collections import deque
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from queue import Empty, Queue
from threading import Thread
from typing import TYPE_CHECKING, Iterator, List

if TYPE_CHECKING:
    from openapi_fal_rest.client import Client

import grpc
import httpx
from httpx_sse import connect_sse
from rich.console import Console
from structlog.typing import EventDict

from fal.api.client import SyncServerlessClient
from fal.sdk import RunnerInfo, RunnerState

from .parser import FalClientParser, SinceAction, get_output_parser


def runners_table(runners: List[RunnerInfo]):
    from rich.table import Table

    table = Table()
    table.add_column("Alias")
    table.add_column("Runner ID")
    table.add_column("In Flight\nRequests")
    table.add_column("Expires In")
    table.add_column("Uptime")
    table.add_column("Revision")
    table.add_column("State")

    for runner in runners:
        external_metadata = runner.external_metadata
        present = external_metadata.get("present_in_group", True)

        num_leases_with_request = len(
            [
                lease
                for lease in external_metadata.get("leases", [])
                if lease.get("request_id") is not None
            ]
        )

        in_flight = str(runner.in_flight_requests)
        missing_leases = runner.in_flight_requests - num_leases_with_request
        if missing_leases > 0:
            # Show a small indicator of in flight requests that are not visible in the
            # leases lists
            # This can be due to race conditions, so only important to report if it's
            # consistent
            in_flight = f"{in_flight} [dim]({missing_leases})[/]"

        uptime = timedelta(
            seconds=int(runner.uptime.total_seconds()),
        )
        table.add_row(
            runner.alias,
            # Mark lost runners in red
            runner.runner_id if present else f"[red]{runner.runner_id}[/]",
            in_flight,
            (
                "N/A"
                if runner.expiration_countdown is None
                else f"{runner.expiration_countdown}s"
            ),
            f"{uptime} ({uptime.total_seconds():.0f}s)",
            runner.revision,
            runner.state.value,
        )

    return table


def runners_requests_table(runners: list[RunnerInfo]):
    from rich.table import Table

    table = Table()
    table.add_column("Runner ID")
    table.add_column("Request ID")
    table.add_column("Caller ID")
    table.add_column("Time Running")

    for runner in runners:
        for lease in runner.external_metadata.get("leases", []):
            if not (req_id := lease.get("request_id")):
                continue

            running_for = None
            try:
                if leased_at_epoch := lease.get("leased_at_epoch"):
                    started_at = datetime.fromtimestamp(
                        float(leased_at_epoch), timezone.utc
                    )
                    running_for = datetime.now(timezone.utc) - started_at
            except Exception:
                pass

            table.add_row(
                runner.runner_id,
                req_id,
                lease.get("caller_user_id") or "",
                str(running_for or ""),
            )

    return table


def _get_tty_size():
    """Get current terminal dimensions."""
    try:
        h, w = struct.unpack("HH", fcntl.ioctl(0, termios.TIOCGWINSZ, b"\0" * 4))[:2]
        return h, w
    except (OSError, ValueError):
        return 24, 80  # Fallback to standard size


def _shell(args):
    """Execute interactive shell in runner."""
    import isolate_proto

    client = SyncServerlessClient(host=args.host, team=args.team)
    stub = client._create_host()._connection.stub
    runner_id = args.id

    # Setup terminal for raw mode
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    tty.setraw(fd)

    # Message queue for stdin data and resize events
    messages = Queue()  # type: ignore
    stop_flag = False

    def handle_resize(*_):
        messages.put(("resize", None))

    signal.signal(signal.SIGWINCH, handle_resize)

    def read_stdin():
        """Read stdin in a background thread."""
        nonlocal stop_flag
        while not stop_flag:
            try:
                data = os.read(fd, 4096)
                if not data:
                    break
                messages.put(("data", data))
            except OSError:
                break

    reader = Thread(target=read_stdin, daemon=True)
    reader.start()

    def stream_inputs():
        """Generate input stream for gRPC."""
        # Send initial message with runner_id
        yield isolate_proto.ShellRunnerInput(runner_id=runner_id)

        # Send terminal size
        msg = isolate_proto.ShellRunnerInput()
        h, w = _get_tty_size()
        msg.tty_size.height = h
        msg.tty_size.width = w
        yield msg

        # Stream stdin data and resize events
        while True:
            try:
                msg_type, data = messages.get(timeout=0.1)
            except Empty:
                continue

            if msg_type == "data":
                yield isolate_proto.ShellRunnerInput(data=data)
            elif msg_type == "resize":
                msg = isolate_proto.ShellRunnerInput()
                h, w = _get_tty_size()
                msg.tty_size.height = h
                msg.tty_size.width = w
                yield msg

    exit_code = 1
    try:
        for output in stub.ShellRunner(stream_inputs()):
            if output.HasField("exit_code"):
                exit_code = output.exit_code
                break
            if output.data:
                sys.stdout.buffer.write(output.data)
                sys.stdout.buffer.flush()
            if output.close:
                break
        exit_code = exit_code or 0
    except grpc.RpcError as exc:
        args.console.print(f"\n[red]Connection error:[/] {exc.details()}")
    except Exception as exc:
        args.console.print(f"\n[red]Error:[/] {exc}")
    finally:
        stop_flag = True
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    return exit_code


def _stop(args):
    client = SyncServerlessClient(host=args.host, team=args.team)
    client.runners.stop(args.id)


def _kill(args):
    client = SyncServerlessClient(host=args.host, team=args.team)
    client.runners.kill(args.id)


def _list_json(args, runners: list[RunnerInfo]):
    json_runners = [
        {
            "alias": r.alias,
            "runner_id": r.runner_id,
            "in_flight_requests": r.in_flight_requests,
            "expiration_countdown": r.expiration_countdown,
            "uptime_seconds": int(r.uptime.total_seconds()),
            "revision": r.revision,
            "state": r.state.value,
        }
        for r in runners
    ]

    res = {
        "runners": json_runners,
    }
    args.console.print(json.dumps(res))


def _list(args):
    client = SyncServerlessClient(host=args.host, team=args.team)
    start_time = args.since
    runners = client.runners.list(since=start_time)

    if args.state:
        states = set(args.state)
        if "all" not in states:
            runners = [
                r
                for r in runners
                if r.state.value.lower() in states
                or (
                    "terminated" in states and r.state.value.lower() == "dead"
                )  # TODO for backwards compatibility. remove later
            ]

    pending_runners = [
        runner for runner in runners if runner.state == RunnerState.PENDING
    ]
    setup_runners = [runner for runner in runners if runner.state == RunnerState.SETUP]
    terminated_runners = [
        runner
        for runner in runners
        if runner.state == RunnerState.DEAD or runner.state == RunnerState.TERMINATED
    ]
    if args.output == "pretty":
        args.console.print(
            "Runners: "
            + str(
                len(runners)
                - len(pending_runners)
                - len(setup_runners)
                - len(terminated_runners)
            )
        )
        args.console.print(f"Runners Pending: {len(pending_runners)}")
        args.console.print(f"Runners Setting Up: {len(setup_runners)}")
        args.console.print(runners_table(runners))

        requests_table = runners_requests_table(runners)
        args.console.print(f"Requests: {len(requests_table.rows)}")
        args.console.print(requests_table)
    elif args.output == "json":
        _list_json(args, runners)
    else:
        raise AssertionError(f"Invalid output format: {args.output}")


def _add_stop_parser(subparsers, parents):
    stop_help = "Stop a runner gracefully."
    parser = subparsers.add_parser(
        "stop",
        description=stop_help,
        help=stop_help,
        parents=parents,
    )
    parser.add_argument(
        "id",
        help="Runner ID.",
    )
    parser.set_defaults(func=_stop)


def _add_kill_parser(subparsers, parents):
    kill_help = "Kill a runner."
    parser = subparsers.add_parser(
        "kill",
        description=kill_help,
        help=kill_help,
        parents=parents,
    )
    parser.add_argument(
        "id",
        help="Runner ID.",
    )
    parser.set_defaults(func=_kill)


def _add_list_parser(subparsers, parents):
    list_help = "List runners."
    parser = subparsers.add_parser(
        "list",
        description=list_help,
        help=list_help,
        parents=[*parents, get_output_parser()],
    )
    parser.add_argument(
        "--since",
        default=None,
        action=SinceAction,
        limit="1 day",
        help=(
            "Show terminated runners since the given time. "
            "Accepts 'now', relative like '30m', '1h', '1d', "
            "or an ISO timestamp. Max 24 hours."
        ),
    )
    parser.add_argument(
        "--state",
        choices=["all", "running", "pending", "setup", "terminated"],
        nargs="+",
        default=None,
        help=("Filter by runner state(s). Choose one or more, or 'all'(default)."),
    )
    parser.set_defaults(func=_list)


def _to_iso_naive(dt: datetime) -> str:
    return dt.astimezone(timezone.utc).isoformat()


def _parse_ts(ts: str) -> datetime:
    # Support both 'Z' and offset formats
    ts_norm = ts.replace("Z", "+00:00")
    return datetime.fromisoformat(ts_norm)


def _to_aware_utc(dt: datetime) -> datetime:
    # Treat naive datetimes as UTC
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def _post_history(
    client: httpx.Client,
    base_params: dict[str, str],
    since: datetime | None,
    until: datetime | None,
    page_size: int,
) -> tuple[list, str | None]:
    params: dict[str, str] = dict(base_params)
    if since is not None:
        params["since"] = _to_iso_naive(since)
    if until is not None:
        params["until"] = _to_iso_naive(until)
    params["page_size"] = str(page_size)
    resp = client.post("/logs/history", params=params)
    if resp.status_code != HTTPStatus.OK:
        try:
            detail = resp.json().get("detail", resp.text)
        except Exception:
            detail = resp.text
        raise RuntimeError(f"Failed to fetch logs history: {detail}")
    data = resp.json()
    items = data.get("items", []) if isinstance(data, dict) else []
    next_until = data.get("next_until") if isinstance(data, dict) else None
    if not isinstance(items, list):
        raise RuntimeError("Unexpected logs history response format")
    return items, next_until


@dataclass
class RestRunnerInfo:
    started_at: datetime | None
    ended_at: datetime | None


def _get_runner_info(rest_client: Client, runner_id: str) -> RestRunnerInfo:
    headers = rest_client.get_headers()
    with httpx.Client(
        base_url=rest_client.base_url, headers=headers, timeout=30
    ) as client:
        resp = client.get(f"/runners/{runner_id}")
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise RuntimeError(f"Runner {runner_id} not found")
        if resp.status_code != HTTPStatus.OK:
            raise RuntimeError(
                f"Failed to fetch runner info: {resp.status_code} {resp.text}"
            )
        data = resp.json()
        if not isinstance(data, dict):
            raise RuntimeError(f"Unexpected runner info response format: {resp.text}")

        start: datetime | None = None
        end: datetime | None = None

        started_at = data.get("started_at")
        if started_at is not None:
            try:
                start = _to_aware_utc(_parse_ts(started_at))
            except Exception:
                start = None

        ended_at = data.get("ended_at")
        if ended_at is not None:
            try:
                end = _to_aware_utc(_parse_ts(ended_at))
            except Exception:
                end = None

        return RestRunnerInfo(started_at=start, ended_at=end)


def _stream_logs(
    rest_client: Client,
    base_params: dict[str, str],
    since: datetime | None,
    until: datetime | None,
) -> Iterator[dict]:
    headers = rest_client.get_headers()
    params: dict[str, str] = base_params.copy()
    if since is not None:
        params["since"] = _to_iso_naive(since)
    if until is not None:
        params["until"] = _to_iso_naive(until)
    with httpx.Client(
        base_url=rest_client.base_url,
        headers=headers,
        timeout=None,
        follow_redirects=True,
    ) as client:
        with connect_sse(
            client,
            method="POST",
            url="/logs/stream",
            params=params,
            headers={"Accept": "text/event-stream"},
        ) as event_source:
            for sse in event_source.iter_sse():
                if not sse.data:
                    continue
                if sse.event == "error":
                    raise RuntimeError(f"Error streaming logs: {sse.data}")
                try:
                    yield json.loads(sse.data)
                except Exception:
                    continue


DEFAULT_PAGE_SIZE = 1000


def _iter_logs(
    rest_client: Client,
    base_params: dict[str, str],
    start: datetime | None,
    end: datetime | None,
) -> Iterator[dict]:
    headers = rest_client.get_headers()
    with httpx.Client(
        base_url=rest_client.base_url,
        headers=headers,
        timeout=300,
        follow_redirects=True,
    ) as client:
        cursor_until = end
        while True:
            items, next_until = _post_history(
                client, base_params, start, cursor_until, DEFAULT_PAGE_SIZE
            )

            yield from items

            if not next_until:
                break

            new_until_dt = _to_aware_utc(_parse_ts(next_until))
            if start is not None and new_until_dt <= start:
                break
            cursor_until = new_until_dt


def _get_logs(
    rest_client: Client,
    params: dict[str, str],
    since: datetime | None,
    until: datetime | None,
    lines_count: int | None,
    *,
    oldest: bool = False,
) -> Iterator[dict]:
    if lines_count is None:
        yield from _iter_logs(rest_client, params, since, until)
        return

    if oldest:
        produced = 0
        for log in _iter_logs(rest_client, params, since, until):
            if produced >= lines_count:
                break
            produced += 1
            yield log
        return

    # newest tail: collect into a fixed-size deque, then yield
    tail: deque[dict] = deque(maxlen=lines_count)
    for log in _iter_logs(rest_client, params, since, until):
        tail.append(log)
    for log in tail:
        yield log


class LogPrinter:
    def __init__(self, console: Console) -> None:
        from structlog.dev import ConsoleRenderer

        from fal.logging.style import LEVEL_STYLES

        self._console = console
        self._renderer = ConsoleRenderer(level_styles=LEVEL_STYLES)

    def _render_log(self, log: dict) -> str:
        ts_str: str = log["timestamp"]
        timestamp = _to_aware_utc(_parse_ts(ts_str))
        local_ts = timestamp.astimezone()
        tz_offset = local_ts.strftime("%z")
        # Insert ':' into offset for readability, e.g. +0300 -> +03:00
        if tz_offset and len(tz_offset) == 5:
            tz_offset = tz_offset[:3] + ":" + tz_offset[3:]

        event: EventDict = {
            "event": log.get("message", ""),
            "level": str(log.get("level", "")).upper(),
            "timestamp": f"{local_ts.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]}{tz_offset}",
        }
        return self._renderer(logger={}, name=event["level"], event_dict=event)

    def print(self, log: dict) -> None:
        self._console.print(self._render_log(log), highlight=False)


DEFAULT_STREAM_SINCE = timedelta(minutes=1)


def _logs(args):
    params: dict[str, str] = {"job_id": args.id}
    if args.search is not None:
        params["search"] = args.search

    client = SyncServerlessClient(host=args.host, team=args.team)
    rest_client = client._create_rest_client()
    runner_info = _get_runner_info(rest_client, args.id)
    follow: bool = args.follow
    since = args.since
    if follow:
        since = since or (datetime.now(timezone.utc) - DEFAULT_STREAM_SINCE)
    else:
        since = since or runner_info.started_at
    until = args.until or runner_info.ended_at

    # Normalize to aware UTC for comparisons
    if since is not None:
        since = _to_aware_utc(since)
    if until is not None:
        until = _to_aware_utc(until)

    # Sanity limiters: clamp within runner lifetime when known
    if runner_info.started_at is not None:
        if since is not None and since < runner_info.started_at:
            since = runner_info.started_at
        if until is not None and until < runner_info.started_at:
            until = runner_info.started_at
    if runner_info.ended_at is not None:
        if since is not None and since > runner_info.ended_at:
            since = runner_info.ended_at
        if until is not None and until > runner_info.ended_at:
            until = runner_info.ended_at

    # Ensure ordering if both are present
    if since is not None and until is not None and until < since:
        since, until = until, since

    lines_arg = args.lines
    lines_count: int | None = None
    lines_oldest = False
    if lines_arg is not None:
        if lines_arg.startswith("+"):
            lines_str = lines_arg[1:]
            lines_oldest = True
        else:
            lines_str = lines_arg
        try:
            lines_count = int(lines_str)
        except ValueError:
            args.parser.error("Invalid -n|--lines value. Use an integer or +integer.")

    if follow:
        logs_gen = _stream_logs(rest_client, params, since, until)
    else:
        logs_gen = _get_logs(
            rest_client, params, since, until, lines_count, oldest=lines_oldest
        )

    printer = LogPrinter(args.console)

    if follow:
        for log in logs_gen:
            if args.output == "json":
                args.console.print(json.dumps(log))
            else:
                printer.print(log)
        return

    if args.output == "json":
        args.console.print(json.dumps({"logs": list(logs_gen)}))
    else:
        for log in reversed(list(logs_gen)):
            printer.print(log)


def _add_logs_parser(subparsers, parents):
    logs_help = "Show logs for a runner."
    parser = subparsers.add_parser(
        "logs",
        aliases=["log"],
        description=logs_help,
        help=logs_help,
        parents=[*parents, get_output_parser()],
    )
    parser.add_argument(
        "id",
        help="Runner ID.",
    )
    parser.add_argument(
        "--search",
        default=None,
        help="Search for string in logs.",
    )
    parser.add_argument(
        "--since",
        default=None,
        action=SinceAction,
        help=(
            "Show logs since the given time. "
            "Accepts 'now', relative like '30m', '1h', or an ISO timestamp. "
            "Defaults to runner start time or to '1m ago' in --follow mode."
        ),
    )
    parser.add_argument(
        "--until",
        default=None,
        action=SinceAction,
        help=(
            "Show logs until the given time. "
            "Accepts 'now', relative like '30m', '1h', or an ISO timestamp. "
            "Defaults to runner finish time or 'now' if it is still running."
        ),
    )
    parser.add_argument(
        "--follow",
        "-f",
        action="store_true",
        help="Follow logs live. If --since is not specified, implies '--since 1m ago'.",
    )
    parser.add_argument(
        "--lines",
        "-n",
        default=None,
        type=str,
        help=(
            "Only show latest N log lines. "
            "If '+' prefix is used, show oldest N log lines. "
            "Ignored if --follow is used."
        ),
    )
    parser.set_defaults(func=_logs)


def _add_shell_parser(subparsers, parents):
    """Add hidden shell command parser."""
    parser = subparsers.add_parser(
        "shell",
        help=argparse.SUPPRESS,
        parents=parents,
    )
    parser.add_argument("id", help="Runner ID.")
    parser.set_defaults(func=_shell)


def add_parser(main_subparsers, parents):
    runners_help = "Manage fal runners."
    parser = main_subparsers.add_parser(
        "runners",
        description=runners_help,
        help=runners_help,
        parents=parents,
        aliases=["machine"],  # backwards compatibility
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        required=True,
        parser_class=FalClientParser,
    )

    _add_stop_parser(subparsers, parents)
    _add_kill_parser(subparsers, parents)
    _add_list_parser(subparsers, parents)
    _add_logs_parser(subparsers, parents)
    _add_shell_parser(subparsers, parents)



import asyncio

import pytest

from fal.distributed.worker import DistributedRunner, DistributedWorker


class SimpleWorker(DistributedWorker):
    def __call__(self, **kwargs):
        return {"status": "ok"}


def test_worker_event_loop():
    """Test that a worker can be created and has a running event loop."""
    worker = SimpleWorker(rank=0, world_size=1)
    assert worker.running
    worker.shutdown()


def test_worker_task_submission():
    """Test that tasks can be submitted to a worker and return results."""
    worker = SimpleWorker(rank=0, world_size=1)
    future = worker.run_in_worker(worker.__call__)
    result = future.result()
    assert result == {"status": "ok"}
    worker.shutdown()


def test_runner_initialization():
    """Test that a distributed runner can be initialized with correct parameters."""
    runner = DistributedRunner(SimpleWorker, world_size=2)
    assert runner.world_size == 2
    assert runner.worker_cls == SimpleWorker
    assert not runner.is_alive()  # Should not be alive before start()


def test_worker_shutdown_is_safe():
    """Test that shutdown doesn't crash even if called multiple times."""
    worker = SimpleWorker(rank=0, world_size=1)
    worker.shutdown()
    worker.shutdown()  # Should not crash


def test_runner_not_alive_before_start():
    """Test that runner reports correctly before starting."""
    runner = DistributedRunner(SimpleWorker, world_size=1)
    assert not runner.is_alive()
    with pytest.raises(RuntimeError, match="not running"):
        runner.ensure_alive()


def test_worker_handles_errors_in_callable():
    """Test that worker properly handles errors raised in __call__."""

    class ErrorWorker(DistributedWorker):
        def error_method(self):
            raise ValueError("Intentional test error")

    worker = ErrorWorker(rank=0, world_size=1)
    future = worker.run_in_worker(worker.error_method)

    with pytest.raises(ValueError, match="Intentional test error"):
        future.result()

    worker.shutdown()


def test_worker_handles_async_functions():
    """Test that worker can handle async functions."""

    class AsyncWorker(DistributedWorker):
        async def async_method(self):
            await asyncio.sleep(0.01)
            return {"async": "result"}

    worker = AsyncWorker(rank=0, world_size=1)
    future = worker.run_in_worker(worker.async_method)
    result = future.result()

    assert result == {"async": "result"}
    worker.shutdown()


def test_worker_submit_when_not_running():
    """Test that submitting to a stopped worker raises error."""
    worker = SimpleWorker(rank=0, world_size=1)
    worker.shutdown()

    async def dummy():
        return {}

    with pytest.raises(RuntimeError, match="Event loop is not running"):
        worker.submit(dummy())


def test_add_streaming_result():
    """Test that streaming results are queued correctly."""
    worker = SimpleWorker(rank=0, world_size=1)
    worker.add_streaming_result({"frame": 1})
    worker.add_streaming_result({"frame": 2})

    # Check queue has items
    assert not worker.queue.empty()
    item1 = worker.queue.get_nowait()
    assert item1 is not None

    worker.shutdown()


def test_add_streaming_error():
    """Test that errors can be added to streaming queue."""
    worker = SimpleWorker(rank=0, world_size=1)
    worker.add_streaming_error(ValueError("test error"))

    assert not worker.queue.empty()
    error_item = worker.queue.get_nowait()
    assert error_item is not None

    worker.shutdown()


def test_runner_gather_errors_when_no_errors():
    """Test gather_errors returns empty list when no errors."""
    runner = DistributedRunner(SimpleWorker, world_size=1)
    errors = runner.gather_errors()
    assert errors == []


def test_runner_zmq_socket_cleanup():
    """Test that ZMQ socket can be safely closed multiple times."""
    runner = DistributedRunner(SimpleWorker, world_size=1)
    assert runner.zmq_socket is None

    runner.close_zmq_socket()  # Should not crash even if None
    assert runner.zmq_socket is None

    runner.close_zmq_socket()  # Should not crash if called again
    assert runner.zmq_socket is None


def test_runner_terminate_when_not_started():
    """Test that terminate() is safe to call even when not started."""
    runner = DistributedRunner(SimpleWorker, world_size=1)
    runner.terminate()  # Should not crash



"""Integration tests for environment management using the API layer."""

from __future__ import annotations

import uuid
from datetime import datetime, timezone

import pytest

from fal.api import FalServerlessError, SyncServerlessClient


@pytest.fixture(scope="function")
def test_env_name():
    """Generate a unique environment name for testing."""
    return f"test-env-{uuid.uuid4().hex[:8]}"


@pytest.fixture(scope="function")
def client():
    """Create a SyncServerlessClient for API layer testing."""
    import os

    host = os.getenv("FAL_HOST")
    api_key = os.getenv("FAL_KEY")
    return SyncServerlessClient(host=host, api_key=api_key)


@pytest.mark.flaky(max_runs=3)
def test_environment_lifecycle(client: SyncServerlessClient, test_env_name: str):
    """Test creating, listing, and deleting environments."""
    # Create environment
    env = client.environments.create(
        test_env_name, description="Integration test environment"
    )
    assert env.name == test_env_name
    assert env.description == "Integration test environment"
    assert env.created_at is not None

    try:
        # List environments - verify our env exists
        environments = client.environments.list()
        env_names = [e.name for e in environments]
        assert test_env_name in env_names

        # Find our environment in the list
        our_env = next(e for e in environments if e.name == test_env_name)
        assert our_env.description == "Integration test environment"

    finally:
        # Cleanup - delete environment
        client.environments.delete(test_env_name)

        # Verify deletion
        environments_after = client.environments.list()
        env_names_after = [e.name for e in environments_after]
        assert test_env_name not in env_names_after


@pytest.mark.flaky(max_runs=3)
def test_environment_with_secrets(client: SyncServerlessClient, test_env_name: str):
    """Test that secrets are scoped to environments."""
    secret_name = f"TEST_SECRET_{uuid.uuid4().hex[:8]}"

    # Create test environment
    client.environments.create(test_env_name, description="Test env for secrets")

    try:
        # Set a secret in the test environment
        client.secrets.set(secret_name, "test_value", environment_name=test_env_name)

        # List secrets in test environment
        secrets_in_env = client.secrets.list(environment_name=test_env_name)
        secret_names = [s.name for s in secrets_in_env]
        assert secret_name in secret_names

        # Verify secret has correct environment
        our_secret = next(s for s in secrets_in_env if s.name == secret_name)
        assert our_secret.environment_name == test_env_name

        # Delete the secret before deleting the environment
        client.secrets.unset(secret_name, environment_name=test_env_name)

        # Delete the environment
        client.environments.delete(test_env_name)

        # Verify the environment is gone
        environments = client.environments.list()
        assert test_env_name not in [e.name for e in environments]

    except Exception:
        # Cleanup on failure
        try:
            client.secrets.unset(secret_name, environment_name=test_env_name)
        except Exception:
            pass
        try:
            client.environments.delete(test_env_name)
        except Exception:
            pass
        raise


def test_delete_nonexistent_environment(client: SyncServerlessClient):
    """Test deleting an environment that doesn't exist raises FalServerlessError."""
    with pytest.raises(FalServerlessError):
        client.environments.delete(f"nonexistent-env-{uuid.uuid4().hex[:8]}")


def test_create_duplicate_environment(client: SyncServerlessClient, test_env_name: str):
    """Test creating an environment with a duplicate name raises FalServerlessError."""
    # Create first environment
    client.environments.create(test_env_name)

    try:
        # Attempt to create duplicate - should raise FalServerlessError
        with pytest.raises(FalServerlessError):
            client.environments.create(test_env_name)
    finally:
        # Cleanup
        client.environments.delete(test_env_name)


@pytest.mark.flaky(max_runs=3)
def test_list_environments_returns_list(client: SyncServerlessClient):
    """Test that list_environments returns a valid list."""
    environments = client.environments.list()
    assert isinstance(environments, list)


@pytest.mark.flaky(max_runs=3)
def test_environment_created_at_timestamp(
    client: SyncServerlessClient, test_env_name: str
):
    """Test that created_at timestamp is properly set."""
    # Create environment
    env = client.environments.create(test_env_name)

    try:
        assert env.created_at is not None
        # Check that it's a datetime object
        assert isinstance(env.created_at, datetime)
        # Check that it's recent (within last minute)
        now = datetime.now(timezone.utc)
        time_diff = (now - env.created_at).total_seconds()
        assert 0 <= time_diff < 60, f"Created time seems wrong: {time_diff}s ago"

    finally:
        client.environments.delete(test_env_name)



from __future__ import annotations

from ._base import (
    AppException,
    AppFileUploadException,
    FalServerlessException,
    FieldException,
    FileTooLargeError,
    RequestCancelledException,
)
from ._cuda import CUDAOutOfMemoryException
from .auth import UnauthenticatedException

__all__ = [
    "FalServerlessException",
    "AppException",
    "FieldException",
    "RequestCancelledException",
    "FileTooLargeError",
    "AppFileUploadException",
    "CUDAOutOfMemoryException",
    "UnauthenticatedException",
]



from fal.api.client import SyncServerlessClient

from .parser import FalClientParser, get_output_parser


def _list(args):
    import json

    client = SyncServerlessClient(host=args.host, team=args.team)
    environments = client.environments.list()

    if args.output == "json":
        json_envs = [
            {
                "name": env.name,
                "description": env.description,
                "is_default": env.is_default,
                "created_at": str(env.created_at),
            }
            for env in environments
        ]
        args.console.print(json.dumps({"environments": json_envs}))
    elif args.output == "pretty":
        from rich.table import Table

        table = Table()
        table.add_column("Name")
        table.add_column("Description")
        table.add_column("Default")
        table.add_column("Created At")

        for env in environments:
            table.add_row(
                env.name,
                env.description or "",
                "Yes" if env.is_default else "",
                str(env.created_at),
            )

        args.console.print(table)
    else:
        raise AssertionError(f"Invalid output format: {args.output}")


def _add_list_parser(subparsers, parents):
    list_help = "List environments."
    parser = subparsers.add_parser(
        "list",
        description=list_help,
        help=list_help,
        parents=[*parents, get_output_parser()],
    )
    parser.set_defaults(func=_list)


def _create(args):
    client = SyncServerlessClient(host=args.host, team=args.team)
    env = client.environments.create(args.name, description=args.description)
    args.console.print(f"Created environment '{env.name}'")


def _add_create_parser(subparsers, parents):
    create_help = "Create an environment."
    parser = subparsers.add_parser(
        "create",
        description=create_help,
        help=create_help,
        parents=parents,
    )
    parser.add_argument(
        "name",
        help="Environment name.",
    )
    parser.add_argument(
        "--description",
        help="Environment description.",
    )
    parser.set_defaults(func=_create)


def _delete(args):
    if not args.yes:
        args.console.print(
            f"[bold yellow]Warning:[/bold yellow] Deleting environment "
            f"'{args.name}' will permanently delete:\n"
            "  • All secrets in this environment\n"
            "  • All apps deployed to this environment\n"
        )
        confirmation = input(
            f"Type the environment name '{args.name}' to confirm deletion: "
        ).strip()

        if confirmation != args.name:
            args.console.print(
                "[red]Deletion cancelled.[/red] Environment name did not match."
            )
            return

    client = SyncServerlessClient(host=args.host, team=args.team)
    client.environments.delete(args.name)
    args.console.print(f"[green]Deleted environment '{args.name}'[/green]")


def _add_delete_parser(subparsers, parents):
    delete_help = "Delete an environment."
    parser = subparsers.add_parser(
        "delete",
        description=delete_help,
        help=delete_help,
        parents=parents,
    )
    parser.add_argument(
        "name",
        help="Environment name.",
    )
    parser.add_argument(
        "--yes",
        action="store_true",
        help="Skip confirmation prompt.",
    )
    parser.set_defaults(func=_delete)


def add_parser(main_subparsers, parents):
    envs_help = "Manage fal environments."
    parser = main_subparsers.add_parser(
        "environments",
        aliases=["envs"],
        parents=parents,
        description=envs_help,
        help=envs_help,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        metavar="command",
        required=True,
        parser_class=FalClientParser,
    )

    _add_list_parser(subparsers, parents)
    _add_create_parser(subparsers, parents)
    _add_delete_parser(subparsers, parents)



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

from fal.api import FAL_SERVERLESS_DEFAULT_URL, FalServerlessHost
from fal.sdk import (
    AliasInfo,
    Credentials,
    EnvironmentInfo,
    KeyScope,
    RunnerInfo,
    ServerlessSecret,
    UserKeyInfo,
)

if TYPE_CHECKING:
    from openapi_fal_rest.client import Client

from . import apps as apps_api
from . import deploy as deploy_api
from . import environments as environments_api
from . import keys as keys_api
from . import runners as runners_api
from . import secrets as secrets_api


class AppsNamespace:
    """Namespace for app management operations.

    Corresponds to `fal apps ...` CLI commands.

    Accessed via `client.apps`.

    Example:
        from fal.api import SyncServerlessClient

        client = SyncServerlessClient()
        apps = client.apps.list()
        client.apps.scale("my-app", max_concurrency=10)
    """

    def __init__(self, client: SyncServerlessClient):
        self.client = client

    def list(
        self, *, filter: str | None = None, environment_name: str | None = None
    ) -> List[AliasInfo]:
        """List all applications. Corresponds to `fal apps list`.

        Args:
            filter: Optional app name filter string.
            environment_name: Optional environment name.

        Example:
            apps = client.apps.list()
            filtered = client.apps.list(filter="stable")
        """
        return apps_api.list_apps(
            self.client, filter=filter, environment_name=environment_name
        )

    def runners(
        self,
        app_name: str,
        *,
        since=None,
        state: List[str] | None = None,
        environment_name: str | None = None,
    ) -> List[RunnerInfo]:
        """List runners for a specific app. Corresponds to `fal apps runners <app>`.

        Args:
            app_name: Name of the application.
            since: Only return runners started after this datetime.
            state: Filter by runner state (e.g., ["running"]).
            environment_name: Optional environment name.

        Example:
            from datetime import datetime, timedelta

            runners = client.apps.runners("my-app")
            recent = client.apps.runners(
                "my-app", since=datetime.now() - timedelta(hours=1)
            )
            running = client.apps.runners("my-app", state=["running"])
        """
        return apps_api.apps_runners(
            self.client,
            app_name,
            since=since,
            state=state,
            environment_name=environment_name,
        )

    def scale(
        self,
        app_name: str,
        *,
        keep_alive: int | None = None,
        max_multiplexing: int | None = None,
        max_concurrency: int | None = None,
        min_concurrency: int | None = None,
        concurrency_buffer: int | None = None,
        concurrency_buffer_perc: int | None = None,
        scaling_delay: int | None = None,
        request_timeout: int | None = None,
        startup_timeout: int | None = None,
        machine_types: List[str] | None = None,
        regions: List[str] | None = None,
        environment_name: str | None = None,
    ) -> apps_api.AliasInfo:
        """Adjust scaling settings for an application. Corresponds to `fal apps scale`.

        Any omitted option keeps the current value.

        Args:
            app_name: Name of the application.
            keep_alive: Keep-alive time in seconds.
            max_multiplexing: Maximum request multiplexing.
            max_concurrency: Maximum concurrent runners.
            min_concurrency: Minimum concurrent runners.
            request_timeout: Request timeout in seconds.
            startup_timeout: Startup timeout in seconds.
            machine_types: List of allowed machine types (e.g., ["GPU-H100"]).
            regions: List of allowed regions (e.g., ["us-east-1"]).

        Example:
            client.apps.scale(
                "my-app",
                keep_alive=300,
                max_concurrency=10,
                min_concurrency=1,
                machine_types=["GPU-H100", "GPU-H200"],
            )
        """
        return apps_api.scale_app(
            self.client,
            app_name,
            keep_alive=keep_alive,
            max_multiplexing=max_multiplexing,
            max_concurrency=max_concurrency,
            min_concurrency=min_concurrency,
            concurrency_buffer=concurrency_buffer,
            concurrency_buffer_perc=concurrency_buffer_perc,
            scaling_delay=scaling_delay,
            request_timeout=request_timeout,
            startup_timeout=startup_timeout,
            machine_types=machine_types,
            regions=regions,
            environment_name=environment_name,
        )

    def rollout(
        self, app_name: str, *, force: bool = False, environment_name: str | None = None
    ) -> None:
        return apps_api.rollout_app(
            self.client, app_name, force=force, environment_name=environment_name
        )


class RunnersNamespace:
    """Namespace for runner management operations.

    Corresponds to `fal runners ...` CLI commands.

    Accessed via `client.runners`.

    Example:
        from fal.api import SyncServerlessClient

        client = SyncServerlessClient()
        runners = client.runners.list()
        client.runners.stop("runner-id")
    """

    def __init__(self, client: SyncServerlessClient):
        self.client = client

    def list(self, *, since=None) -> List[RunnerInfo]:
        """List all runners. Corresponds to `fal runners list`.

        Args:
            since: Only return runners started after this datetime.

        Example:
            from datetime import datetime, timedelta

            all_runners = client.runners.list()
            recent = client.runners.list(since=datetime.now() - timedelta(minutes=10))
        """
        return runners_api.list_runners(self.client, since=since)

    def stop(self, runner_id: str) -> None:
        """Gracefully stop a runner.

        Args:
            runner_id: The ID of the runner to stop.
        """
        return runners_api.stop_runner(self.client, runner_id)

    def kill(self, runner_id: str) -> None:
        """Forcefully kill a runner.

        Args:
            runner_id: The ID of the runner to kill.
        """
        return runners_api.kill_runner(self.client, runner_id)


class KeysNamespace:
    """Namespace for API key management. Corresponds to `fal keys ...` CLI commands.

    Accessed via `client.keys`.

    Example:
        from fal.api import SyncServerlessClient

        client = SyncServerlessClient()
        keys = client.keys.list()
        key_id, key_secret = client.keys.create(scope="admin")
    """

    def __init__(self, client: SyncServerlessClient):
        self.client = client

    def create(
        self, *, scope: KeyScope, description: str | None = None
    ) -> tuple[str, str]:
        """Create a new API key.

        Args:
            scope: Key scope (e.g., "admin").
            description: Optional description for the key.

        Returns:
            Tuple of (key_id, key_secret).
        """
        return keys_api.create_key(self.client, scope=scope, description=description)

    def list(self) -> List[UserKeyInfo]:
        """List all API keys."""
        return keys_api.list_keys(self.client)

    def revoke(self, key_id: str) -> None:
        """Revoke an API key.

        Args:
            key_id: The ID of the key to revoke.
        """
        return keys_api.revoke_key(self.client, key_id)


class SecretsNamespace:
    """Namespace for secrets management. Corresponds to `fal secrets ...` CLI commands.

    Accessed via `client.secrets`.

    Example:
        from fal.api import SyncServerlessClient

        client = SyncServerlessClient()
        client.secrets.set("API_KEY", "my-secret-value")
        secrets = client.secrets.list()
    """

    def __init__(self, client: SyncServerlessClient):
        self.client = client

    def set(self, name: str, value: str, environment_name: str | None = None) -> None:
        """Set a secret value.

        Args:
            name: Name of the secret.
            value: Value to store.
        """
        return secrets_api.set_secret(
            self.client, name, value, environment_name=environment_name
        )

    def list(self, environment_name: str | None = None) -> List[ServerlessSecret]:
        """List all secrets (names only, not values)."""
        return secrets_api.list_secrets(self.client, environment_name=environment_name)

    def unset(self, name: str, environment_name: str | None = None) -> None:
        """Delete a secret.

        Args:
            name: Name of the secret to delete.
            environment_name: Optional environment name.
        """
        return secrets_api.unset_secret(
            self.client, name, environment_name=environment_name
        )


class _EnvironmentsNamespace:
    def __init__(self, client: SyncServerlessClient):
        self.client = client

    def create(self, name: str, description: str | None = None) -> EnvironmentInfo:
        return environments_api.create_environment(
            self.client, name, description=description
        )

    def list(self) -> List[EnvironmentInfo]:
        return environments_api.list_environments(self.client)

    def delete(self, name: str) -> None:
        return environments_api.delete_environment(self.client, name)


@dataclass
class SyncServerlessClient:
    """Synchronous Python client for fal Serverless.

    Manage apps, runners, and deployments programmatically. The namespaces
    and methods mirror the CLI so you can automate the same workflows from Python.

    Args:
        host: Optional. Override API host.
        api_key: Optional. If omitted, read from env/profile.
        profile: Optional. Named profile to use.
        team: Optional. Team context for runner operations.

    Example:
        from fal.api import SyncServerlessClient

        client = SyncServerlessClient()

        # List apps
        apps = client.apps.list()

        # Scale an app
        client.apps.scale("my-app", max_concurrency=10)

        # Deploy
        client.deploy("path/to/myfile.py::MyApp")

    Namespaces:
        - client.apps.*    - corresponds to `fal apps ...`
        - client.runners.* - corresponds to `fal runners ...`
        - client.keys.*    - corresponds to `fal keys ...`
        - client.secrets.* - corresponds to `fal secrets ...`
        - client.deploy()  - corresponds to `fal deploy ...`
    """

    host: Optional[str] = None
    api_key: Optional[str] = None
    profile: Optional[str] = None
    team: Optional[str] = None

    def __post_init__(self) -> None:
        self.apps = AppsNamespace(self)
        self.runners = RunnersNamespace(self)
        self.keys = KeysNamespace(self)
        self.secrets = SecretsNamespace(self)
        self.environments = _EnvironmentsNamespace(self)

    def deploy(self, *args, **kwargs):
        """Deploy an application. Corresponds to `fal deploy`.

        If app_ref is omitted, discovery behavior matches the CLI
        (e.g., uses pyproject.toml).

        Args:
            app_ref: Path to file, file::Class, or existing app name.
            app_name: Override the application name.
            auth: Authentication mode ("private" | "public").
            strategy: Deployment strategy ("recreate" | "rolling").
            reset_scale: If False, use previous scaling settings.

        Example:
            # Auto-discover from pyproject.toml
            client.deploy()

            # Deploy from a file path
            client.deploy("path/to/myfile.py")

            # Deploy a specific class
            client.deploy("path/to/myfile.py::MyApp")

            # With options
            client.deploy(
                app_ref="path/to/myfile.py::MyApp",
                app_name="myapp",
                auth="public",
                strategy="rolling",
            )
        """
        return deploy_api.deploy(self, *args, **kwargs)

    # Internal helpers
    @property
    def _grpc_host(self) -> str:
        return self.host or FAL_SERVERLESS_DEFAULT_URL

    @property
    def _rest_url(self) -> str:
        from fal.flags import REST_SCHEME

        return f"{REST_SCHEME}://{self._grpc_host.replace('api', 'rest', 1)}"

    @property
    def _credentials(self) -> Credentials:
        from fal.sdk import FalServerlessKeyCredentials, get_default_credentials

        if self.api_key:
            if self.team:
                raise ValueError(
                    "Using explicit team with key credentials is not allowed"
                )
            try:
                key_id, key_secret = self.api_key.split(":", 1)
            except ValueError:
                raise ValueError("api_key must be in 'KEY_ID:KEY_SECRET' format")
            return FalServerlessKeyCredentials(key_id, key_secret)

        if self.profile:
            prev = os.environ.get("FAL_PROFILE")
            os.environ["FAL_PROFILE"] = self.profile
            try:
                return get_default_credentials(team=self.team)
            finally:
                if prev is None:
                    os.environ.pop("FAL_PROFILE", None)
                else:
                    os.environ["FAL_PROFILE"] = prev

        return get_default_credentials(team=self.team)

    def _create_host(
        self, *, local_file_path: str = "", environment_name: str | None = None
    ) -> FalServerlessHost:
        return FalServerlessHost(
            self._grpc_host,
            local_file_path=local_file_path,
            credentials=self._credentials,
            environment_name=environment_name,
        )

    def _create_rest_client(self) -> Client:
        from openapi_fal_rest.client import Client

        import fal.flags as flags

        return Client(
            self._rest_url,
            headers=self._credentials.to_headers(),
            timeout=30,
            verify_ssl=not flags.TEST_MODE,
            raise_on_unexpected_status=False,
            follow_redirects=True,
        )



from fal.cli.main import parse_args
from fal.cli.secrets import _list, _set, _unset


def test_set():
    args = parse_args(["secrets", "set", "secret1=value1", "secret2=value2"])
    assert args.func == _set
    assert args.secrets == {"secret1": "value1", "secret2": "value2"}


def test_set_with_env():
    args = parse_args(["secrets", "set", "secret1=value1", "--env", "dev"])
    assert args.func == _set
    assert args.secrets == {"secret1": "value1"}
    assert args.env == "dev"


def test_list():
    args = parse_args(["secrets", "list"])
    assert args.func == _list


def test_list_with_env():
    args = parse_args(["secrets", "list", "--env", "prod"])
    assert args.func == _list
    assert args.env == "prod"


def test_unset():
    args = parse_args(["secrets", "unset", "secret"])
    assert args.func == _unset
    assert args.secret == "secret"


def test_unset_with_env():
    args = parse_args(["secrets", "unset", "secret", "--env", "staging"])
    assert args.func == _unset
    assert args.secret == "secret"
    assert args.env == "staging"



from fal.cli.environments import _create, _delete, _list
from fal.cli.main import parse_args


def test_list():
    args = parse_args(["environments", "list"])
    assert args.func == _list


def test_list_alias():
    args = parse_args(["envs", "list"])
    assert args.func == _list


def test_list_json_output():
    args = parse_args(["environments", "list", "--output", "json"])
    assert args.func == _list
    assert args.output == "json"


def test_create():
    args = parse_args(["environments", "create", "dev"])
    assert args.func == _create
    assert args.name == "dev"


def test_create_with_description():
    args = parse_args(
        ["environments", "create", "dev", "--description", "Development environment"]
    )
    assert args.func == _create
    assert args.name == "dev"
    assert args.description == "Development environment"


def test_create_alias():
    args = parse_args(["envs", "create", "staging"])
    assert args.func == _create
    assert args.name == "staging"


def test_delete():
    args = parse_args(["environments", "delete", "dev"])
    assert args.func == _delete
    assert args.name == "dev"


def test_delete_alias():
    args = parse_args(["envs", "delete", "staging"])
    assert args.func == _delete
    assert args.name == "staging"


def test_delete_has_yes_flag():
    """Test that the delete command accepts --yes flag"""
    args = parse_args(["environments", "delete", "dev", "--yes"])
    assert args.func == _delete
    assert args.name == "dev"
    assert args.yes is True


def test_delete_without_yes_flag():
    """Test that the delete command works without --yes flag (requires confirmation)"""
    args = parse_args(["environments", "delete", "dev"])
    assert args.func == _delete
    assert args.name == "dev"
    assert args.yes is False


def test_delete_alias_with_yes():
    """Test that the delete alias accepts --yes flag"""
    args = parse_args(["envs", "delete", "staging", "--yes"])
    assert args.func == _delete
    assert args.name == "staging"
    assert args.yes is True



from __future__ import annotations

import hashlib
import os
import zipfile
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from openapi_fal_rest.client import Client

from pathspec import PathSpec


def _check_hash(client: Client, target_path: str, hash_string: str) -> bool:
    import openapi_fal_rest.api.files.check_dir_hash as check_dir_hash_api
    import openapi_fal_rest.models.hash_check as hash_check_model

    response = check_dir_hash_api.sync_detailed(
        target_path,
        client=client,
        json_body=hash_check_model.HashCheck(hash_string),
    )

    res: bool = response.parsed  # type: ignore
    return response.status_code == 200 and res


def _upload_file(
    client: Client, source_path: str, target_path: str, unzip: bool = False
):
    import openapi_fal_rest.api.files.upload_local_file as upload_local_file_api
    import openapi_fal_rest.models.body_upload_local_file as upload_file_model
    import openapi_fal_rest.types as rest_types

    with open(source_path, "rb") as file_to_upload:
        body = upload_file_model.BodyUploadLocalFile(
            rest_types.File(
                payload=file_to_upload,
                # We need to set a file_name, otherwise the server errors
                # processing the file
                file_name=os.path.basename(source_path),
            )
        )

        response = upload_local_file_api.sync_detailed(
            target_path,
            client=client,
            unzip=unzip,
            multipart_data=body,
        )

    if response.status_code != 200:
        raise Exception(
            "Failed to upload file. "
            "Server returned status code "
            f"{response.status_code} and message {response.parsed}"
        )


def _compute_directory_hash(dir_path: str) -> str:
    hash = hashlib.sha256()
    for root, _, files in os.walk(dir_path):
        for file in files:
            file_path = os.path.join(root, file)
            if file != ".fal_hash":
                with open(file_path, "rb") as f:
                    hash.update(f.read())
    return hash.hexdigest()


def _load_gitignore_patterns(dir_path: str) -> list:
    # TODO: consider looking at .gitignore files in child directories
    gitignore_path = os.path.join(dir_path, ".gitignore")
    if os.path.exists(gitignore_path):
        with open(gitignore_path) as f:
            gitignore_patterns = f.read().splitlines()
    else:
        gitignore_patterns = []
    return gitignore_patterns


def _is_ignored(file_path: str, gitignore_patterns: list[str]) -> bool:
    pathspec = PathSpec.from_lines("gitwildmatch", gitignore_patterns)
    return pathspec.match_file(file_path)


def _zip_directory(dir_path: str, zip_path: str) -> None:
    gitignore_patterns = _load_gitignore_patterns(dir_path)

    with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
        for root, _, files in os.walk(dir_path):
            for file in files:
                file_path = os.path.join(root, file)
                relative_path = os.path.relpath(file_path, dir_path)

                if not _is_ignored(relative_path, gitignore_patterns):
                    arcname = relative_path
                    zipf.write(file_path, arcname)


def sync_dir(local_dir: str | Path, remote_dir: str, force_upload=False) -> str:
    from fal.api.client import SyncServerlessClient

    local_dir_abs = os.path.expanduser(local_dir)
    if not os.path.isabs(remote_dir) or not remote_dir.startswith("/data"):
        raise ValueError(
            "'remote_dir' must be an absolute path starting with `/data`, "
            "e.g. '/data/sync/my_dir'"
        )

    remote_dir = remote_dir.replace("/data/", "", 1)

    # Compute the local directory hash
    local_hash = _compute_directory_hash(local_dir_abs)

    client = SyncServerlessClient()._create_rest_client()

    print(f"Syncing {local_dir} with {remote_dir}...")

    if _check_hash(client, remote_dir, local_hash) and not force_upload:
        print(f"{remote_dir} already uploaded and matches {local_dir}")
        return remote_dir

    with open(os.path.join(local_dir_abs, ".fal_hash"), "w") as f:
        f.write(local_hash)

    # Zip the local directory
    zip_path = f"{local_dir_abs}.zip"

    _zip_directory(local_dir_abs, zip_path)

    # Upload the zipped directory to the serverless environment
    _upload_file(client, zip_path, remote_dir, unzip=True)

    os.remove(zip_path)

    print("Done")

    # Return the full path to the remote directory
    return remote_dir



from unittest import mock

import pytest

from fal.toolkit.utils.retry import retry


def test_retry_success():
    @retry(max_retries=3, base_delay=1, backoff_type="fixed")
    def successful_function():
        return "Success"

    assert successful_function() == "Success"


def test_retry_failure():
    @retry(max_retries=3, base_delay=1, backoff_type="fixed")
    def failing_function():
        raise ValueError("Test error")

    with pytest.raises(ValueError, match="Test error"):
        failing_function()


@mock.patch("time.sleep")
def test_retry_exponential(mock_sleep):
    attempts = 0

    @retry(max_retries=3, base_delay=1, max_delay=60, backoff_type="exponential")
    def fail_twice_then_succeed():
        nonlocal attempts
        attempts += 1
        if attempts < 3:
            raise ValueError("Test error")
        return "Success"

    assert fail_twice_then_succeed() == "Success"
    assert attempts == 3
    mock_sleep.assert_has_calls(
        [
            mock.call(1),
            mock.call(2),
        ]
    )


@mock.patch("time.sleep")
def test_retry_fixed(mock_sleep):
    attempts = 0

    @retry(max_retries=3, base_delay=2, backoff_type="fixed")
    def fail_twice_then_succeed():
        nonlocal attempts
        attempts += 1
        if attempts < 3:
            raise ValueError("Test error")
        return "Success"

    assert fail_twice_then_succeed() == "Success"
    assert attempts == 3
    mock_sleep.assert_has_calls(
        [
            mock.call(2),
            mock.call(2),
        ]
    )


@mock.patch("random.uniform", return_value=1.25)
@mock.patch("time.sleep")
def test_retry_jitter(mock_sleep, mock_uniform):
    attempts = 0

    @retry(max_retries=3, base_delay=1, backoff_type="fixed", jitter=True)
    def fail_twice_then_succeed():
        nonlocal attempts
        attempts += 1
        if attempts < 3:
            raise ValueError("Test error")
        return "Success"

    assert fail_twice_then_succeed() == "Success"
    assert attempts == 3
    mock_sleep.assert_has_calls(
        [
            mock.call(1.25),
            mock.call(1.25),
        ]
    )



import re

import rich

# = or := only
KV_SPLIT_RE = re.compile(r"(=|:=)")


def _api(args):
    """Handle the api command execution."""
    from . import cli_nested_json

    params_split = [KV_SPLIT_RE.split(param) for param in args.params]
    params = cli_nested_json.interpret_nested_json(  # type: ignore
        [(key, value) for key, _, value in params_split]
    )

    if args.model_id.endswith("/stream"):
        stream_run(args.model_id, params)
    else:
        queue_run(args.model_id, params)


def stream_run(model_id: str, params: dict):
    import fal.apps

    res = fal.apps.stream(model_id, params)  # type: ignore
    for line in res:
        if isinstance(line, str):
            rich.print(line)
        else:
            if isinstance(line, memoryview):
                rich.print(line.tobytes().decode())
            else:
                rich.print(line.decode())


def queue_run(model_id: str, params: dict):
    from rich.console import Group
    from rich.live import Live
    from rich.panel import Panel
    from rich.text import Text

    import fal.apps

    handle = fal.apps.submit(model_id, params)  # type: ignore
    logs = []  # type: ignore

    with Live(auto_refresh=False) as live:
        for event in handle.iter_events(logs=True):
            if isinstance(event, fal.apps.Queued):
                status = Text(f"⏳ Queued (position: {event.position})", style="yellow")
            elif isinstance(event, fal.apps.InProgress):
                status = Text("🔄 In Progress", style="blue")
                if event.logs:
                    logs.extend(log.get("message", str(log)) for log in event.logs)
                    logs = logs[-10:]  # Keep only last 10 logs
            else:
                status = Text("✅ Done", style="green")

            status_panel = Panel(status, title="Status")
            logs_panel = Panel("\n".join(logs), title="Logs")

            live.update(Group(status_panel, logs_panel))
            live.refresh()

        # Show final result
        result = handle.get()
        live.update(rich.pretty.Pretty(result))


def add_parser(main_subparsers, parents):
    """Add the api command to the main parser."""
    api_help = "Call a fal API endpoint directly"
    parser = main_subparsers.add_parser(
        "api",
        description=api_help,
        help=api_help,
        parents=parents,
    )

    parser.add_argument(
        "model_id",
        help="Name of the Model ID to call",
    )

    parser.add_argument(
        "params",
        nargs="*",
        help="Key-value pairs (e.g. key=value or nested[a][b]=value)",
    )

    parser.set_defaults(func=_api)



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from fal.app import App

# Global reference to the currently running app
# This is useful for accessing the app in runtime
current_app: Optional["App"] = None  # type: ignore[assignment]


def set_current_app(app: "App"):
    global current_app  # noqa: PLW0603
    assert current_app is None, "current_app is already set"
    current_app = app


def get_current_app() -> Optional["App"]:
    return current_app



from __future__ import annotations

from fal.toolkit.audio.audio import Audio, AudioField
from fal.toolkit.compilation import (
    get_gpu_type,
    load_inductor_cache,
    sync_inductor_cache,
    synchronized_inductor_cache,
)
from fal.toolkit.exceptions import (
    FalTookitException,
    FileUploadException,
    KVStoreException,
)
from fal.toolkit.file import CompressedFile, File, FileField
from fal.toolkit.image.image import Image, ImageField, ImageSizeInput, get_image_size
from fal.toolkit.kv import KVStore
from fal.toolkit.optimize import optimize
from fal.toolkit.utils import (
    FAL_MODEL_WEIGHTS_DIR,
    FAL_PERSISTENT_DIR,
    FAL_REPOSITORY_DIR,
    clone_repository,
    download_file,
    download_model_weights,
)
from fal.toolkit.utils.download_utils import DownloadError
from fal.toolkit.video.video import Video, VideoField

__all__ = [
    # Classes
    "Audio",
    "AudioField",
    "CompressedFile",
    "File",
    "FileField",
    "Image",
    "ImageField",
    "ImageSizeInput",
    "KVStore",
    "Video",
    "VideoField",
    # Functions
    "get_image_size",
    "optimize",
    "clone_repository",
    "download_file",
    "download_model_weights",
    "get_gpu_type",
    "load_inductor_cache",
    "sync_inductor_cache",
    "synchronized_inductor_cache",
    # Constants
    "FAL_MODEL_WEIGHTS_DIR",
    "FAL_PERSISTENT_DIR",
    "FAL_REPOSITORY_DIR",
    # Exceptions
    "FalTookitException",
    "FileUploadException",
    "KVStoreException",
    "DownloadError",
]



class FalTookitException(Exception):
    """Base exception for all toolkit exceptions"""

    pass


class FileUploadException(FalTookitException):
    """Raised when file upload fails"""

    pass


class KVStoreException(FalTookitException):
    """Raised when KV store operation fails"""

    pass
